/*
   This file is part of GNOME Tetravex.

   Copyright (C) 2019 Arnaud Bonatti

   GNOME Tetravex is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 2 of the License, or
   (at your option) any later version.

   GNOME Tetravex is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this GNOME Tetravex.  If not, see <https://www.gnu.org/licenses/>.
*/

              .score-title {
  font-weight:bolder;

  font-size:200%;
  margin-bottom:1.5rem;
  margin-top:0.5rem;

  transition:font-size      0.5s,
             margin-bottom  0.5s,
             margin-top     0.5s;
}
.small-window .score-title {
  font-size:150%;
  margin-bottom:1rem;
  margin-top:0.2rem;
}

grid.score-overlay button {
  margin-top:0.5rem;
}

.bold-label {
  font-weight:bolder;
}

.italic-label {
  font-style:italic;
}

button.rounded-button {
  border-radius:       9999px;
  -gtk-outline-radius: 9999px;
}

/*\
* * Scores dialog
\*/

.score-dialog treeview:dir(ltr) {
  padding-left: 0.3rem;
  padding-right: 0.6rem;
}
.score-dialog treeview:dir(rtl) {
  padding-left: 0.6rem;
  padding-right: 0.3rem;
}

.score-dialog treeview > header > button {
  padding-left:0.6rem;
  padding-right:0.6rem;
}

/*\
* * the end
\*/
